;;;======================
;;; TONAL ANALYSIS IN OM
;;; Using Java objects
;;;======================

(in-package :ha)

;;; TESTS
;(setf factory (new OMAnalysisFactory.))
;(setf analyzer (OMAnalysisFactory.create "Harmonic"))
;(OMAnalysis.getDescription analyzer)
;(setf segmentation (OMHarmonicAnalysis.getSegmentation analyzer))
#|
(loop while (Segmentation.hasNext segmentation) do
      (setf segment (Segmentation.next segmentation))
      (print (unbox-int (Segment.get segment "t1")))
      (print (unbox-int (Segment.get segment "t2")))
      (print (unbox-string (Segment.get segment "segment-data")))
)
(Segmentation.size segmentation)
(Segmentation.first segmentation)
|#

(defmethod! midip-Crel (pitches)
  (mapcar #'(lambda (x) (mod x 12)) pitches)
)

(defmethod! midic-Crel (pitches)
  (midip-Crel (mapcar #'(lambda (x) (/ x 100)) pitches))
)

(defmethod analysis-as-chordseq ((object t))
  (let ((data (osc-harmonic-analysis-as-chordseq object)))
    (make-instance 'om::chord-seq
                   :lmidic (nth 0 data)
                   :lonset (nth 1 data)
                   :ldur (nth 2 data))
    )
)

(om::defclass* om::chord-seq-analysis (chord-seq)
  ((analysis :initform nil :accessor analysis :initarg :analysis :type om::OM-ABSTRACT-ANALYSIS)))

(defmethod om::chord-seq-analysis ((self om::chord-seq-analysis))
  (om::analyse-object self))

(defmethod initialize-instance :after ((self om::chord-seq-analysis) &rest initargs  &key (Empty nil) (LPort nil))
  (om::set-object-analysis self (om::slot-value self 'analysis))
  (om::run-analysis self))

;;;======================
;;; GROUNDTRUTH ANALYSIS
;;;======================

(om::defclass! GROUNDTRUTH-ANALYSIS (om::OM-ABSTRACT-ANALYSIS)
   ((source :accessor source :initarg :source :initform "" :documentation "source file")))

;(defmethod analysis-segments ((self GROUNDTRUTH-ANALYSIS)) (om::analysis-segments self))
;(defmethod source ((self om::GROUNDTRUTH-ANALYSIS)) (om::source self))

(defmethod segment-data-tostring ((self GROUNDTRUTH-ANALYSIS) segment) (om::segment-data segment))

(defmethod om::compute-analysis-segments ((self GROUNDTRUTH-ANALYSIS) (object om::chord-seq))
  (loop for chord in (analysis-from-textfile (source self))
        collect
        (make-instance 'om::marker-segment
                       :mrk-time (second chord)
                       :segment-data (first chord)
                       :color *om-gray-color*
        )))

(defmethod om::draw-segment-data ((self GROUNDTRUTH-ANALYSIS) segment view) 
  (oa::om-with-font oa::*om-default-font2b*
     (om::om-with-fg-color view (om::color segment)
         (oa::om-draw-string (+ 
                              (om::get-key-space view)
                              (om::ms2pixel (om::mrk-time segment) (/ (om::staff-size view) 4) (om::staff-zoom view)) 
                              6)
                             20
                             (om::segment-data segment)))))


;;;========================
;;; PARDO MINIMAL SEGMENTS
;;;========================

(om::defclass! PARDO-MINSEG-ANALYSIS (om::OM-ABSTRACT-ANALYSIS)
  ()
  (:icon 242)
  (:documentation ""))

(defmethod segment-data-tostring ((self PARDO-MINSEG-ANALYSIS) segment) "")

(defmethod om::compute-analysis-segments ((self PARDO-MINSEG-ANALYSIS) (object om::chord-seq))
  (loop for chord in (om::real-chords object)
        collect
        (make-instance 'om::marker-segment
                       :mrk-time (om::offset->ms chord)
                       :segment-data nil
                       :color om::*om-red2-color*))
)

(defmethod om::draw-segment-data ((self PARDO-MINSEG-ANALYSIS) segment view) 
  nil)

;;;===================
;;; HARMONIC ANALYSIS
;;;===================

;-------------------
; HARMONIC SEGMENTS
;-------------------
(defstruct ha-chord
  root
  type)

(defclass! harmonic-segment (om::segment)
  ((id :accessor id :initarg :id :initform 0 :documentation "segment ID")
   (onset :accessor onset :initarg :onset :initform 0 :documentation "onet time (ms)")
   (offset :accessor offset :initarg :offset :initform 0 :documentation "offset time (ms)")
   (ppoints :accessor ppoints :initarg :ppoints :initform nil :documentation "partition points")
   (parent :accessor parent :initarg :parent :initform nil :documentation "Harmonic analysis")
   (active :accessor active :initarg :active :initform nil :documentation "")
   (erasable :accessor erasable :initarg :erasable :initform t :documentation "")
   (previous :accessor previous :initarg :previous :initform nil :documentation "")
   (next :accessor next :initarg :next :initform nil :documentation "")
   ; rendering options
   (min-width :accessor min-width :initarg :min-width :initform 0 :documentation "minimum width required to display the information")))

(defmethod om::segment-begin ((self harmonic-segment)) (onset self))
; TODO: compute segment length
(defmethod om::segment-end ((self harmonic-segment)) (offset self))

(defmethod om::draw-segment ((self harmonic-segment) view) 
  (let* ((font oa::*om-default-font2b*)
         (color (om::color self))
         (size (get-note-size view))
         (btn-size 16)
         (staff-top (get-staff-top-px view))
         (staff-h (get-staff-height-px view))
         (staff-line-h (get-staff-line-height-px view))
         (line-top (- staff-top (* 8 staff-line-h)))
         (line-bottom (+ (+ staff-top staff-h) (* 12 staff-line-h)))
         (xonset (get-onset-in-px (onset self) view))
         (xoffset (- (get-onset-in-px (offset self) view) 2))
         (em (oa::om-string-size "m" font))
         (btn-x (+ xonset em))
         (btn-y (- staff-top (* 6 staff-line-h)))
         ; If the previous segment overlaps, this segment is drawn a line down
         (prev (previous self))
         (line (if prev (if (> (+ (get-onset-in-px (onset prev) view) (min-width prev)) xonset) 1 0) 0))
         (label (concatenate 'string (om::segment-data-tostring (parent self) self) " (" (when prev (write-to-string (min-width prev))) ")"))
         (label-x (+ xonset 8))
         ; Label ypos is 6 lines from the staff
         (label-y (+ line-bottom -4 (* line 20)));(- staff-top (* 6 staff-line-h) 4))
         (label-w (oa::om-string-size label font))
         (label-h (oa::om-string-h font)))

    ; Draws the label
    (oa::om-with-font font                   
       (oa::om-with-fg-color view color
         (oa::om-draw-string (+ label-x 18) label-y label)))

    ; Draws the segment
    (oa::om-with-line-size 1
      (oa::om-with-line :dash
        (oa::om-with-fg-color view om::*om-steel-blue-color*
          (oa::om-draw-line xonset line-top xonset line-bottom))
          (when (active self)
          #|(oa::om-with-fg-color view om::*om-red2-color*
            (oa::om-draw-line xonset (+ (+ staff-top staff-h) (* 8 line-h)) xoffset (+ (+ staff-top staff-h) (* 8 line-h))))|#
            (om-with-hilite-foreground
              (om-fill-rect xonset line-top (- xoffset xonset) (- line-bottom line-top)))
        )))

    ; Draws the partition points
    (when (active self)
      (om-with-fg-color view om::*om-red2-color*
        (om-with-line-size 1
          (om-with-dashline '(3 6)
            (loop for ppoint in (ppoints self)
                  do
                  (let ((ppoint-x (get-onset-in-px ppoint view))
                        (ppoint-ytop (- staff-top (* 2 staff-line-h)))
                        (ppoint-ybottom (+ (+ staff-top staff-h) (* 4 staff-line-h))))
                    (om-draw-line ppoint-x ppoint-ytop ppoint-x ppoint-ybottom)
                    (om-draw-picture view (om::om-load-and-store-picture "add" 'harmonic-analysis) 
                                     (om-make-point (- ppoint-x 7) (+ ppoint-ybottom 4)) (om-make-point btn-size btn-size))))
            ))))

    ; If the previous segment is selected this segment can be merged
    (when (previous self)
      (when (active (previous self))
        (om-draw-picture view (om::om-load-and-store-picture "merge" 'harmonic-analysis) 
                         (om-make-point (+ 4 xonset) (+ 4 (+ staff-top staff-h) (* 4 staff-line-h))) (om-make-point btn-size btn-size))
        ))

    ; Draws the play button
    (om-draw-picture view (om::om-load-and-store-picture "sound" 'harmonic-analysis) 
                     (om-make-point label-x (- label-y 13)) (om-make-point btn-size btn-size))
      
    (when (active self)
      ; Draws the accept button
      (om-draw-picture view (om::om-load-and-store-picture "tick" 'harmonic-analysis) 
                       (om-make-point (+ btn-x 18) btn-y) (om-make-point btn-size btn-size))

      ; Draws the edit button
      (om-draw-picture view (om::om-load-and-store-picture "pencil" 'harmonic-analysis) 
                       (om-make-point (+ btn-x 36) btn-y) (om-make-point btn-size btn-size)))

      ; Draws the delete button
      (when (erasable self)
        (om-draw-picture view (om::om-load-and-store-picture "delete" 'harmonic-analysis) 
                         (om-make-point (+ btn-x 54) btn-y) (om-make-point btn-size btn-size)))
    ))

;--------------------
; CANDIDATE SEGMENTS
;--------------------
(defclass! harmonic-candidate-segment (om::segment)
  ((onset :accessor onset :initarg :onset :initform 0 :documentation "onet time (ms)")))

(defmethod om::segment-begin ((self harmonic-candidate-segment)) (onset self))
; TODO: compute segment length
(defmethod om::segment-end ((self harmonic-candidate-segment)) (+ 1000 (onset self)))

(defmethod om::draw-segment ((self harmonic-candidate-segment) view) 
  (let* ((font oa::*om-default-font2b*)
         (color (om::color self))
         (size (get-note-size view))
         (pos-x (- (get-onset-in-px (segment-begin self) view) size))
         (pos-ytop (- (get-staff-top-px view) (* 2 (get-staff-line-height-px view))))
         (pos-ybottom (+ (+ (get-staff-top-px view) (get-staff-height-px view)) (* 8 (get-staff-line-height-px view)))))

    ; Draws a dashed line
    #|(oa::om-with-fg-color view om::*om-red2-color*
      (oa::om-with-line-size 1
        (oa::om-with-line :dash
          (oa::om-draw-line line-x line-ytop line-x line-ybottom)
          (oa::om-draw-line (- line-x size) line-ytop (+ line-x size) line-ytop)
          (oa::om-draw-line (- line-x size) line-ybottom (+ line-x size) line-ybottom)
        )))|#
    (oa::om-with-fg-color view color
      (om-with-font font
         (oa::om-draw-string pos-x pos-ybottom "^")))

    ; Draws the create button
    #|(om-draw-picture view (om::om-load-and-store-picture "add" 'harmonic-analysis) 
                     (om-make-point btn-x btn-y) (om-make-point btn-size btn-size))|#

    ))

;------------


(om::defclass! HARMONIC-ANALYSIS (om::OM-ABSTRACT-ANALYSIS)
   ((id :accessor id :initarg :id :initform nil :documentation "")
    (roots :accessor roots :initarg :roots :initform nil :documentation "")
    (types :accessor types :initarg :types :initform '(("maj" (0 4 37)) ("min" (0 3 7))) :documentation "")
    (onsets :accessor onsets :initarg :onsets :initform nil :documentation "list of onset times")
    (active :accessor active :initarg :active :initform nil :documentation "active segment"))
   (:icon 242)
   (:documentation ""))

;(defmethod analysis-segments ((self HARMONIC-ANALYSIS)) (om::analysis-segments self))

(defmethod om::segment-data-tostring ((self HARMONIC-ANALYSIS) (seg om::segment)) (om::segment-data seg))
(defmethod om::segment-data-tostring ((self HARMONIC-ANALYSIS) (seg HARMONIC-SEGMENT))
  (let* ((data (om::segment-data seg))
         (root (root-label self (ha-chord-root data)))
         (type (ha-chord-type data)))
    (concatenate 'string root type)))

(defmethod initialize-instance :after ((self HARMONIC-ANALYSIS) &rest args)
  (setf (onsets self)
        (mapcar 'segment-begin (om::slot-value self 'onsets)))
  self)

(defmethod type-intervals ((self HARMONIC-ANALYSIS) type)
  (let ((data (find type (types self) :key #'first :test #'equal)))
    (if data
        (second data)
      nil)))

(defmethod root-label ((self HARMONIC-ANALYSIS) root)
  (let ((data (find root (roots self) :key #'first)))
    (if data
        (second data)
      nil)))

(defmethod om::compute-analysis-segments ((self HARMONIC-ANALYSIS) (object om::chord-seq))
  (let* ((response (if (onsets self)
                       (osc-harmonic-analysis-with-onsets self object (list (onsets self)))
                     (osc-harmonic-analysis self object))))
    (when response
      (setf (id self) (first response)))
    (setf (roots self) (first (second response)))
    (setf (types self) (second (second response)))
    (build-segments self (third (second response)))
    ))

(defmethod build-segments ((self HARMONIC-ANALYSIS) (data list))
  (let* ((isfirst t)
         (seg nil)
         (segments nil))
    (setf segments
          (loop for item in data
                collect
                (setf seg
                      (make-instance 'harmonic-segment
                                     :id (third item)
                                     :onset (fourth item)
                                     :offset (fifth item)
                                     :ppoints (sixth item)
                                     :parent self
                                     :active (if isfirst (progn (setf isfirst nil) t) nil)
                                     :erasable nil
                                     :previous seg
                                     :segment-data (make-ha-chord :root (first item) :type (second item))
                                     :color om::*om-steel-blue-color*
                                     :min-width (+ 26 (om-string-size (first item) om::*om-default-font2b*))
                                     ))))
    ; segments is a double-linked list
    ; this step sets the forward links
    (setf seg nil)
    (loop for item in (reverse segments)
          do
          (setf (next item) seg)
          (setf seg item))
    segments))



(defmethod change-label ((self HARMONIC-ANALYSIS) seg-id root type)
  (let ((data (osc-harmonic-analysis-change-label (id self) seg-id root type)))
    (if data
        (progn
          (setf (analysis-segments self)
                (build-segments self (third data)))
          t)
      nil)))

; Draws the segment labels in the view
(defmethod om::draw-segment-data ((self HARMONIC-ANALYSIS) segment view)
  )

;;;===================
;;; TONAL ANALYSIS
;;;===================

(om::defclass! TONAL-ANALYSIS (om::OM-ABSTRACT-ANALYSIS)
   ((onsets :accessor onsets :initarg :onsets :initform nil :documentation "list of onset times"))
   (:icon 242)
   (:documentation ""))

(defmethod om::segment-data-tostring ((self TONAL-ANALYSIS) segment) (om::segment-data segment))

(defmethod initialize-instance :after ((self TONAL-ANALYSIS) &rest args)
  (setf (onsets self)
        (mapcar 'segment-begin (onsets self)))
  self)

(defmethod om::compute-analysis-segments ((self TONAL-ANALYSIS) (object om::voice))
  (loop for item in (osc-tonal-analysis object)
        collect
        (make-instance 'om::time-segment
                       :t1 (nth 0 item)
                       :t2 (nth 0 item)
                       :segment-data (nth 1 item)
                       :color om::*om-steel-blue-color*
                       )
        )
)
